@@ -21,8 +21,6 @@ module IntelliMonad.Tools
   ( module IntelliMonad.Tools.Utils,
     module IntelliMonad.Tools.Arxiv,
     module IntelliMonad.Tools.Bash,
-    module IntelliMonad.Tools.TextToSpeech,
-    module IntelliMonad.Tools.DallE3,
     defaultTools,
   )
 where
@@ -30,8 +28,6 @@ where
 import Data.Proxy
 import IntelliMonad.Tools.Arxiv
 import IntelliMonad.Tools.Bash
-import IntelliMonad.Tools.DallE3
-import IntelliMonad.Tools.TextToSpeech
 import IntelliMonad.Tools.Utils
 import IntelliMonad.Types
 
@@ -39,10 +35,6 @@ arxiv = ToolProxy (Proxy :: Proxy Arxiv)
 
 bash = ToolProxy (Proxy :: Proxy Bash)
 
-textToSpeech = ToolProxy (Proxy :: Proxy TextToSpeech)
-
-dallE3 = ToolProxy (Proxy :: Proxy DallE3)
-
 defaultTools :: [ToolProxy]
 defaultTools =
   [ bash,
