@@ -40,7 +40,9 @@ import Data.Map (Map)
 import qualified Data.Aeson as A
 import qualified Data.Aeson.Key as A
 import qualified Data.Aeson.KeyMap as A
+import qualified Data.Aeson.KeyMap as HM
 import qualified Data.Aeson.Text as A
+import Data.List (nub)
 import Data.ByteString (ByteString, fromStrict, toStrict)
 import Data.Coerce
 import Data.Kind (Type)
@@ -148,6 +150,13 @@ class HasFunctionObject r where
   getFunctionDescription :: String
   getFieldDescription :: String -> String
 
+-- | Constructor schema for sum types
+data ConstructorSchema = ConstructorSchema
+  { csName :: Text           -- ^ Constructor name
+  , csPayload :: Schema      -- ^ Payload schema
+  , csIsNullary :: Bool      -- ^ True for zero-field constructors
+  } deriving (Show, Eq)
+
 data Schema
   = Maybe' Schema
   | String'
@@ -157,6 +166,11 @@ data Schema
   | Array' Schema
   | Boolean'
   | Null'
+  -- Sum type schemas
+  | Enum' [Text]                       -- ^ String enum for nullary constructors
+  | OneOfUntagged [ConstructorSchema]  -- ^ Untagged union (distinguishable shapes)
+  | OneOfTagged [ConstructorSchema]    -- ^ Tagged union with @tag/@value
+  deriving (Show, Eq)
 
 class GSchema s f where
   gschema :: forall a. f a -> Schema
@@ -166,51 +180,9 @@ class JSONSchema r where
   default schema :: (HasFunctionObject r, Generic r, GSchema r (Rep r)) => Schema
   schema = gschema @r (from (undefined :: r))
 
-data OpenAI
-
-data LLMProtocol
-  = OpenAI
-
-data LLMRequest
-  = LouterRequest Louter.ChatRequest
-  deriving (Show, Eq, Generic)
-
-data LLMResponse
-  = LouterResponse Louter.ChatResponse
-  deriving (Show, Eq, Generic)
-
-data LLMTool
-  = LouterTool Louter.Tool
-  deriving (Show, Eq, Generic)
-
-instance ToJSON LLMRequest
-instance FromJSON LLMRequest
-instance ToJSON LLMResponse
-instance FromJSON LLMResponse
-
--- Manual Ord instances (Louter types don't have Ord)
-instance Ord LLMRequest where
-  compare _ _ = EQ  -- Simplified: treat all requests as equal for ordering
-
-instance Ord LLMResponse where
-  compare _ _ = EQ  -- Simplified: treat all responses as equal for ordering
-
-instance Ord LLMTool where
-  compare _ _ = EQ  -- Simplified: treat all tools as equal for ordering
-
-class LLMApi api where
-  type LLMRequest' api
-  type LLMResponse' api
-  type LLMTool' api
-  defaultRequest :: LLMRequest' api
-  newTool :: forall a. (HasFunctionObject a, JSONSchema a) => Proxy a -> LLMTool' api
-  toTools :: LLMRequest' api -> [LLMTool' api]
-  fromTools :: LLMRequest' api -> [LLMTool' api] -> LLMRequest' api
-  fromModel_ :: Text -> LLMRequest' api
-
 class ChatCompletion b where
-  toRequest :: LLMRequest -> b -> LLMRequest
-  fromResponse :: Text -> LLMResponse -> (b, FinishReason)
+  toRequest :: Louter.ChatRequest -> b -> Louter.ChatRequest
+  fromResponse :: Text -> Louter.ChatResponse -> (b, FinishReason)
 
 toPV :: (ToJSON a) => a -> PersistValue
 toPV = toPersistValue . toStrict . encode
@@ -250,20 +222,6 @@ instance PersistField Message where
 instance PersistFieldSql Message where
   sqlType _ = sqlType (Proxy @ByteString)
 
-instance PersistField LLMRequest where
-  toPersistValue = toPV
-  fromPersistValue = fromPV
-
-instance PersistField LLMResponse where
-  toPersistValue = toPV
-  fromPersistValue = fromPV
-
-instance PersistFieldSql LLMRequest where
-  sqlType _ = sqlType (Proxy @ByteString)
-
-instance PersistFieldSql LLMResponse where
-  sqlType _ = sqlType (Proxy @ByteString)
-
 
 
 share
@@ -281,8 +239,8 @@ Content
     deriving FromJSON
     deriving Generic
 Context
-    request LLMRequest
-    response LLMResponse Maybe
+    request Louter.ChatRequest
+    response Louter.ChatResponse Maybe
     header [Content]
     body [Content]
     footer [Content]
@@ -291,7 +249,6 @@ Context
     created UTCTime default=CURRENT_TIME
     deriving Show
     deriving Eq
-    deriving Ord
 KeyValue
     namespace Text
     key Text
@@ -302,6 +259,15 @@ KeyValue
     deriving Ord
 |]
 
+-- Manual Ord instance for Context
+-- We compare only the fields that have Ord, ignoring request and response
+instance Ord Context where
+  compare c1 c2 =
+    compare (contextHeader c1, contextBody c1, contextFooter c1,
+             contextTotalTokens c1, contextSessionName c1, contextCreated c1)
+            (contextHeader c2, contextBody c2, contextFooter c2,
+             contextTotalTokens c2, contextSessionName c2, contextCreated c2)
+
 data ToolProxy = forall t. (Tool t, A.FromJSON t, A.ToJSON t, A.FromJSON (Output t), A.ToJSON (Output t), HasFunctionObject t, JSONSchema t) => ToolProxy (Proxy t)
 
 class CustomInstruction a where
@@ -390,6 +356,77 @@ toAeson = \case
       ]
   Boolean' -> A.Object [("type", "boolean")]
   Null' -> A.Object [("type", "null")]
+  -- Sum type schemas
+  Enum' constructors ->
+    A.Object
+      [ ("type", "string"),
+        ("enum", A.Array $ V.fromList $ map A.String constructors)
+      ]
+  OneOfUntagged constructors ->
+    A.Object
+      [ ("oneOf", A.Array $ V.fromList $ map constructorToUntaggedSchema constructors)
+      ]
+  OneOfTagged constructors ->
+    A.Object
+      [ ("oneOf", A.Array $ V.fromList $ map constructorToTaggedSchema constructors)
+      ]
+
+-- | Convert a constructor schema to an untagged JSON schema
+constructorToUntaggedSchema :: ConstructorSchema -> A.Value
+constructorToUntaggedSchema (ConstructorSchema name payload _) =
+  -- For untagged, just emit the payload schema
+  -- The constructor name is used for documentation but not in the schema itself
+  toAeson payload
+
+-- | Convert a constructor schema to a tagged JSON schema with @tag/@value
+constructorToTaggedSchema :: ConstructorSchema -> A.Value
+constructorToTaggedSchema (ConstructorSchema name payload isNullary) =
+  if isNullary
+    then
+      -- Nullary constructor: {"@tag": "ConstructorName"}
+      A.Object
+        [ ("type", "object"),
+          ("properties", A.Object
+            [ ("@tag", A.Object [("type", "string"), ("const", A.String name)])
+            ]),
+          ("required", A.Array $ V.fromList [A.String "@tag"]),
+          ("additionalProperties", A.Bool False)
+        ]
+    else
+      case payload of
+        Object' _ ->
+          -- Object payload: flat tagged format {"@tag": "Ctor", "field1": ..., "field2": ...}
+          case toAeson payload of
+            A.Object payloadObj ->
+              let tagProp = ("@tag", A.Object [("type", "string"), ("const", A.String name)])
+                  -- Extract properties from payload
+                  props = case HM.lookup "properties" payloadObj of
+                    Just (A.Object p) -> p
+                    _ -> HM.empty
+                  -- Extract required fields
+                  req = case HM.lookup "required" payloadObj of
+                    Just (A.Array r) -> V.toList r
+                    _ -> []
+                  -- Merge @tag with payload properties
+                  allProps = HM.insert "@tag" (A.Object [("type", "string"), ("const", A.String name)]) props
+                  allRequired = A.String "@tag" : req
+              in A.Object
+                  [ ("type", "object"),
+                    ("properties", A.Object allProps),
+                    ("required", A.Array $ V.fromList allRequired)
+                  ]
+            _ -> A.Null  -- Shouldn't happen
+        _ ->
+          -- Non-object payload: nested format {"@tag": "Ctor", "@value": ...}
+          A.Object
+            [ ("type", "object"),
+              ("properties", A.Object
+                [ ("@tag", A.Object [("type", "string"), ("const", A.String name)]),
+                  ("@value", toAeson payload)
+                ]),
+              ("required", A.Array $ V.fromList [A.String "@tag", A.String "@value"]),
+              ("additionalProperties", A.Bool False)
+            ]
 
 instance Semigroup Schema where
   (<>) (Object' a) (Object' b) = Object' (a <> b)
@@ -428,7 +465,66 @@ instance (JSONSchema a) => JSONSchema [a] where
 instance JSONSchema () where
   schema = Null'
 
-instance (HasFunctionObject s, JSONSchema c) => GSchema s U1 where
+-- | Helper functions for sum type schema generation
+
+-- | Internal wrapper to track constructor schemas during generic traversal
+-- This allows us to distinguish between a single constructor and a sum type
+data SchemaOrConstructors
+  = SingleSchema Schema  -- ^ Not a sum type, just a regular schema
+  | Constructors [ConstructorSchema]  -- ^ Sum type with multiple constructors
+  deriving (Show, Eq)
+
+-- | Extract constructors from a schema, wrapping non-sum schemas as single-constructor lists
+extractConstructors :: Schema -> [ConstructorSchema]
+extractConstructors (Enum' names) = map (\n -> ConstructorSchema n Null' True) names
+extractConstructors (OneOfUntagged cs) = cs
+extractConstructors (OneOfTagged cs) = cs
+extractConstructors other = [ConstructorSchema "" other (isNullarySchema other)]
+
+-- | Normalize constructor name (e.g., "Red" -> "red")
+normalizeConstructorName :: Text -> Text
+normalizeConstructorName = T.toLower
+
+-- | Check if a schema represents a nullary constructor
+isNullarySchema :: Schema -> Bool
+isNullarySchema Null' = True
+isNullarySchema _ = False
+
+-- | Check if all constructors are nullary (enum pattern)
+isEnum :: [ConstructorSchema] -> Bool
+isEnum = all csIsNullary
+
+-- | Extract schema shape for distinguishability check
+-- Two schemas are distinguishable if they have different shapes
+schemaShape :: Schema -> Text
+schemaShape String' = "string"
+schemaShape Number' = "number"
+schemaShape Integer' = "integer"
+schemaShape Boolean' = "boolean"
+schemaShape Null' = "null"
+schemaShape (Array' _) = "array"
+schemaShape (Object' fields) = "object:" <> T.intercalate "," (map (\(n,_,_) -> T.pack n) fields)
+schemaShape (Maybe' s) = "maybe:" <> schemaShape s
+schemaShape (Enum' _) = "enum"
+schemaShape (OneOfUntagged _) = "oneof-untagged"
+schemaShape (OneOfTagged _) = "oneof-tagged"
+
+-- | Check if constructor shapes are mutually exclusive (distinguishable)
+areShapesDistinguishable :: [ConstructorSchema] -> Bool
+areShapesDistinguishable constructors =
+  let shapes = map (schemaShape . csPayload) constructors
+      uniqueShapes = nub shapes
+  in length shapes == length uniqueShapes
+
+-- | Choose appropriate sum type encoding based on constructor analysis
+chooseSumEncoding :: [ConstructorSchema] -> Schema
+chooseSumEncoding constructors
+  | null constructors = Null'  -- Shouldn't happen, but handle gracefully
+  | isEnum constructors = Enum' (map csName constructors)
+  | areShapesDistinguishable constructors = OneOfUntagged constructors
+  | otherwise = OneOfTagged constructors
+
+instance (HasFunctionObject s) => GSchema s U1 where
   gschema _ = Null'
 
 instance (HasFunctionObject s, JSONSchema c) => GSchema s (K1 i c) where
@@ -437,17 +533,39 @@ instance (HasFunctionObject s, JSONSchema c) => GSchema s (K1 i c) where
 instance (HasFunctionObject s, GSchema s a, GSchema s b) => GSchema s (a :*: b) where
   gschema _ = gschema @s @a undefined <> gschema @s @b undefined
 
+-- | Sum type instance - collects all constructors from both branches
 instance (HasFunctionObject s, GSchema s a, GSchema s b) => GSchema s (a :+: b) where
-  gschema _ = gschema @s @a undefined
-  gschema _ = gschema @s @b undefined
+  gschema _ =
+    let leftSchema = gschema @s @a undefined
+        rightSchema = gschema @s @b undefined
+        leftConstructors = extractConstructors leftSchema
+        rightConstructors = extractConstructors rightSchema
+        allConstructors = leftConstructors ++ rightConstructors
+    in chooseSumEncoding allConstructors
 
--- | Datatype
+-- | Datatype - unwraps single-constructor types to their payload
 instance (HasFunctionObject s, GSchema s f) => GSchema s (M1 D c f) where
-  gschema _ = gschema @s @f undefined
+  gschema _ =
+    let innerSchema = gschema @s @f undefined
+    in case innerSchema of
+         -- Single constructor case: unwrap to just the payload
+         Enum' [_] -> Null'  -- Single nullary constructor is just Null
+         OneOfUntagged [ConstructorSchema _ payload _] -> payload
+         -- Multiple constructors: keep as-is
+         _ -> innerSchema
 
--- | Constructor Metadata
+-- | Constructor Metadata - captures constructor name and wraps payload
 instance (HasFunctionObject s, GSchema s f, Constructor c) => GSchema s (M1 C c f) where
-  gschema _ = gschema @s @f undefined
+  gschema proxy =
+    let name = T.pack $ conName (undefined :: M1 C c f p)
+        normalizedName = normalizeConstructorName name
+        payload = gschema @s @f undefined
+        isNullary = isNullarySchema payload
+    -- Return a single-constructor "sum type" that will be collected by :+:
+    -- or used directly if there's only one constructor
+    in case payload of
+         Null' -> Enum' [normalizedName]  -- Nullary constructor
+         _ -> OneOfUntagged [ConstructorSchema normalizedName payload isNullary]
 
 -- | Selector Metadata
 instance (HasFunctionObject s, GSchema s f, Selector c) => GSchema s (M1 S c f) where
@@ -456,13 +574,13 @@ instance (HasFunctionObject s, GSchema s f, Selector c) => GSchema s (M1 S c f)
         desc = getFieldDescription @s name
      in Object' [(name, desc, (gschema @s @f undefined))]
 
-toolAdd :: forall api a. (LLMApi api, Tool a, HasFunctionObject a, JSONSchema a) => LLMRequest' api -> LLMRequest' api
+toolAdd :: forall a. (Tool a, HasFunctionObject a, JSONSchema a) => Louter.ChatRequest -> Louter.ChatRequest
 toolAdd req =
-  let prevTools = case toTools @api req of
+  let prevTools = case toTools req of
         [] -> []
         v -> v
-      newTools = prevTools ++ [newTool @api @a Proxy]
-   in fromTools @api req newTools
+      newTools = prevTools ++ [newTool @a Proxy]
+   in fromTools req newTools
 
 defaultUTCTime :: UTCTime
 defaultUTCTime = UTCTime (coerce (0 :: Integer)) 0
@@ -530,32 +648,37 @@ class PersistentBackend p where
   deleteKey :: (MonadIO m, MonadFail m) => Conn p -> Unique KeyValue -> m ()
 
 
-instance LLMApi OpenAI where
-  type LLMRequest' OpenAI = Louter.ChatRequest
-  type LLMResponse' OpenAI = Louter.ChatResponse
-  type LLMTool' OpenAI = Louter.Tool
-  defaultRequest =
-    Louter.ChatRequest
-      { Louter.reqModel = "gpt-4"
-      , Louter.reqMessages = []
-      , Louter.reqTools = []
-      , Louter.reqToolChoice = Louter.ToolChoiceAuto
-      , Louter.reqTemperature = Nothing
-      , Louter.reqMaxTokens = Nothing
-      , Louter.reqStream = False
-      }
-  newTool (Proxy :: Proxy a) =
-    Louter.Tool
-      { Louter.toolName = T.pack $ getFunctionName @a
-      , Louter.toolDescription = Just (T.pack $ getFunctionDescription @a)
-      , Louter.toolParameters = toAeson (schema @a)
-      }
-  toTools req = Louter.reqTools req
-  fromTools req tools = req { Louter.reqTools = tools }
-  fromModel_ model =
-    (defaultRequest @OpenAI :: LLMRequest' OpenAI)
-      { Louter.reqModel = model
-      }
+defaultRequest :: Louter.ChatRequest
+defaultRequest =
+  Louter.ChatRequest
+    { Louter.reqModel = "gpt-4"
+    , Louter.reqMessages = []
+    , Louter.reqTools = []
+    , Louter.reqToolChoice = Louter.ToolChoiceAuto
+    , Louter.reqTemperature = Nothing
+    , Louter.reqMaxTokens = Nothing
+    , Louter.reqStream = False
+    }
+
+newTool :: forall a. (HasFunctionObject a, JSONSchema a) => Proxy a -> Louter.Tool
+newTool (Proxy :: Proxy a) =
+  Louter.Tool
+    { Louter.toolName = T.pack $ getFunctionName @a
+    , Louter.toolDescription = Just (T.pack $ getFunctionDescription @a)
+    , Louter.toolParameters = toAeson (schema @a)
+    }
+
+toTools :: Louter.ChatRequest -> [Louter.Tool]
+toTools req = Louter.reqTools req
+
+fromTools :: Louter.ChatRequest -> [Louter.Tool] -> Louter.ChatRequest
+fromTools req tools = req { Louter.reqTools = tools }
+
+fromModel_ :: Text -> Louter.ChatRequest
+fromModel_ model =
+  (defaultRequest :: Louter.ChatRequest)
+    { Louter.reqModel = model
+    }
 
 -- | Read the JSON object and convert it to a Map
 toMap :: Text -> Map Text Value
@@ -569,7 +692,7 @@ fromMap :: Map Text Value -> Text
 fromMap txt = TL.toStrict $ A.encodeToLazyText txt
 
 instance ChatCompletion Contents where
-  toRequest (LouterRequest orgRequest) contents =
+  toRequest orgRequest contents =
     let messages = flip map contents $ \case
           Content user (Message message) _ _ ->
             Louter.Message
@@ -600,9 +723,9 @@ instance ChatCompletion Contents where
               { Louter.msgRole = Louter.RoleTool
               , Louter.msgContent = [Louter.TextPart ret']
               }
-     in LouterRequest $ orgRequest { Louter.reqMessages = messages }
+     in orgRequest { Louter.reqMessages = messages }
 
-  fromResponse sessionName (LouterResponse response) =
+  fromResponse sessionName response =
     let choice = head (Louter.respChoices response)
         message = Louter.choiceMessage choice
         toolCalls = Louter.choiceToolCalls choice
@@ -624,41 +747,21 @@ instance ChatCompletion Contents where
      in (contents, finishReason)
 
     
-data LLMProxy api =
-  LLMProxy
-  { toRequest' :: LLMRequest' api -> LLMRequest
-  , toResponse' :: LLMResponse' api -> LLMResponse
-  }
-
-withLLMRequest
-  :: forall a. LLMRequest
-  -> (forall (api :: Type)
-      . (LLMApi api)
-      => LLMProxy api
-      -> LLMRequest' api
-      -> a)
-  -> a
-withLLMRequest req func =
-  case req of
-    LouterRequest req' -> func (LLMProxy @OpenAI LouterRequest LouterResponse) req'
-
-updateRequest :: LLMRequest -> Contents -> LLMRequest
+updateRequest :: Louter.ChatRequest -> Contents -> Louter.ChatRequest
 updateRequest = toRequest
 
---  withLLMRequest req' $ \(p :: LLMProxy api) req -> (toRequest' @api p) (toRequest req cs)
-
-addTools :: [ToolProxy] -> LLMRequest -> LLMRequest
+addTools :: [ToolProxy] -> Louter.ChatRequest -> Louter.ChatRequest
 addTools [] req' = req'
 addTools (tool : tools') req' =
   case tool of
     (ToolProxy (_ :: Proxy a)) ->
-      withLLMRequest req' $ \(p :: LLMProxy api) req -> addTools tools' ((toRequest' @api p) $ toolAdd @api @a req)
-      
-fromModel :: Text -> LLMRequest
-fromModel = LouterRequest . fromModel_ @OpenAI
+      addTools tools' (toolAdd @a req')
 
-runRequest :: forall a. (ChatCompletion a) => Text -> LLMRequest -> a -> IO ((a, FinishReason), LLMResponse)
-runRequest sessionName (LouterRequest defaultReq) request = do
+fromModel :: Text -> Louter.ChatRequest
+fromModel = fromModel_
+
+runRequest :: forall a. (ChatCompletion a) => Text -> Louter.ChatRequest -> a -> IO ((a, FinishReason), Louter.ChatResponse)
+runRequest sessionName defaultReq request = do
   config <- readConfig
 
   -- Determine backend type (default to OpenAI if not specified)
@@ -684,7 +787,7 @@ runRequest sessionName (LouterRequest defaultReq) request = do
           }
 
   client <- Louter.newClient louterBackend
-  let (LouterRequest req) = (toRequest (LouterRequest defaultReq) request)
+  let req = toRequest defaultReq request
 
   lookupEnv "OPENAI_DEBUG" >>= \case
     Just "1" -> do
@@ -696,4 +799,4 @@ runRequest sessionName (LouterRequest defaultReq) request = do
   result <- Louter.chatCompletion client req
   case result of
     Left err -> error $ T.unpack $ "Louter error: " <> err
-    Right res -> return $ (fromResponse sessionName (LouterResponse res), LouterResponse res)
+    Right res -> return $ (fromResponse sessionName res, res)
